'''
Created on 11.09.2012
@author: ByteHangar.com
$Revision: 17 $
'''
#revision = "$Revision: 17 $"
#rev_no = filter(str.isdigit, revision)

bl_info = {
    'name': 'UV island copy paste',
    'author': 'ByteHangar.com',
    'version': (0, 2, 0),
    'blender': (2, 6, 3),
    'api': 46461,
    "location": "UV/Image editor > Menu > UV island copy paste",
    'description': 'Copy selected UV island and paste to another island',
    'warning': '',
    'wiki_url': '',
    'tracker_url': '',
    "support": 'COMMUNITY',
    "category": "Mesh"}

import bpy
from igraph import Graph
import time

class Timer:
	''' Small class to measure run time of critical parts'''
	def __enter__(self):
		self.start = time.clock()
		return self

	def __exit__(self, *args):
		self.end = time.clock()
		self.interval = self.end - self.start
		
	def delta(self):
		return "{0:.6f} sec".format(self.interval)

class ICP_copy(bpy.types.Operator):
	''' Copy selected UV island to memory '''
	bl_idname = "uv.island_copy"
	bl_label = "Island copy"
	bl_description = "Copy selected UV island to memory"
	
	@classmethod
	def poll(cls, context):
		return (context.object
			and context.mode == "EDIT_MESH"
			and not context.tool_settings.use_uv_select_sync
			and context.tool_settings.uv_select_mode == "ISLAND"
			and context.object.data.uv_layers.active)
	
	def execute(self, context):
		bpy.ops.object.mode_set(mode='OBJECT')
		
		me = bpy.context.object.data
		uvl = me.uv_layers.active.data
		
		UV_loops_selected = [loop.index for loop in me.loops if uvl[loop.index].select]

		bpy.ops.object.mode_set(mode='EDIT')
		
		# !!! need improved check for island selection
		if len(UV_loops_selected) >= 3:
			# store selected in mesh data property
			me["UV_stored_loops"] = UV_loops_selected
			self.report({"INFO"}, "Island data stored")
			return {"FINISHED"}
		else:
			# clear previously selected and warn user
			if me.get("UV_stored_loops", None):
				del me["UV_stored_loops"]
			self.report({"WARNING"}, "Select an island first")
			return {"CANCELLED"}
	
class ICP_paste(bpy.types.Operator):
	''' Paste previously stored UV island to selected '''
	bl_idname = "uv.island_paste"
	bl_label = "Island paste"
	bl_description = "Paste previously stored UV island to selected"
	
	@classmethod
	def poll(cls, context):
		return (context.object
			and context.mode == "EDIT_MESH"
			and not context.tool_settings.use_uv_select_sync
			and context.tool_settings.uv_select_mode == "ISLAND"
			and context.object.data.get("UV_stored_loops", None) )

	def report_and_cancel(self, text):
		# restore EDIT mode
		bpy.ops.object.mode_set(mode='EDIT')
		self.report({"WARNING"}, text)
		return {"CANCELLED"}
	
	def vertex_adj(self, context, vertices):
		'''	Return dictionaries with vertex index as key and sets of adjacent vertex as value
			vertices - vertex index list
		'''
		saved_mode = bpy.context.object.mode
		bpy.ops.object.mode_set(mode='OBJECT')
		
		me = bpy.context.object.data
		
		# prepare dictionaries with vertex index as key and empty set as value
		v_adj = dict((loop.vertex_index, set()) for loop in me.loops if loop.index in vertices)
		# fill in adjacency
		for loop in me.loops:
			if loop.index in vertices:
				v_adj[me.edges[loop.edge_index].vertices[0]].add(me.edges[loop.edge_index].vertices[1])
				v_adj[me.edges[loop.edge_index].vertices[1]].add(me.edges[loop.edge_index].vertices[0])
		
		bpy.ops.object.mode_set(mode=saved_mode)
		return v_adj
	
	def edge_list(self, v_adj):
		''' Return list of tuples, each tuple is edge in graph 
			v_adj - vertex adjacency dict
		'''
		v_list = [vertex for vertex in sorted(v_adj)]
		edge_set = set()
		for vertex, neighbors in v_adj.items():
			for neighbor in neighbors:
				v_index = v_list.index(vertex)
				n_index = v_list.index(neighbor) 
				edge_set.add(tuple(sorted((v_index, n_index))))
		return sorted(edge_set)
	
	def compare(self, v_adj_A, v_adj_B):
		''' Return True if graphs are isomorphic and permutation list if first is true,
			else False and None
			v_adj_* - adjacency dictionaries for both graphs
		'''
		g_A = Graph(self.edge_list(v_adj_A))
		g_B = Graph(self.edge_list(v_adj_B))
		
		with Timer() as t:
			status, mapping_AB, mapping_BA = g_A.isomorphic_bliss(g_B, return_mapping_12=True, return_mapping_21=True)
		print("Isomorpism search done in: ", t.delta())
		mapping = None
		if status:
			v_list = [vertex for vertex in sorted(v_adj_A)]
			mapping = [v_list[index] for index in mapping_BA]
		return status, mapping
	
	def execute(self, context):
		bpy.ops.object.mode_set(mode='OBJECT')
		
		me = bpy.context.object.data
		uvl = me.uv_layers.active.data
		
		self.UV_loops_selected = [loop.index for loop in me.loops if uvl[loop.index].select]
		self.UV_loops_stored = list(context.object.data.get("UV_stored_loops"))

		if self.UV_loops_selected == self.UV_loops_stored:
			return self.report_and_cancel("Can't paste onto itself")
		if len(self.UV_loops_selected) != len(self.UV_loops_stored):
			return self.report_and_cancel("Selected and stored islands has different loops length")

		# prepare vertex adjacency dictionaries
		v_adj_stored = self.vertex_adj(context, self.UV_loops_stored)
		v_adj_selected = self.vertex_adj(context, self.UV_loops_selected)
		
		success, v_mapping = self.compare(v_adj_stored, v_adj_selected)
		
		if success:
			v_selected = v_adj_selected.keys() 
			with Timer() as t:
				self.copy_uv_coords(v_mapping, v_selected)
			print("Copy done in: ", t.delta())
			self.report({"INFO"}, "Island copied")
			bpy.ops.object.mode_set(mode='EDIT')
			return {"FINISHED"}
		else:
			return self.report_and_cancel("Isomorphic variant doesn't found. Islands are different.")
			
	def copy_uv_coords(self, from_vertices, to_vertices):
		bpy.ops.object.mode_set(mode='OBJECT')
			
		me = bpy.context.object.data
		uvl = me.uv_layers.active.data
		
		#prepare dictionary of matching vertex indices
		match = dict(zip(from_vertices, to_vertices))

		#create dictionary with vertex_index as key and tuple of UV-coords as value for UV_loops_selected
		uv_coords = dict( (me.loops[loop_index].vertex_index, (uvl[loop_index].uv[0], uvl[loop_index].uv[1])) for loop_index in self.UV_loops_selected)
		
		# update uv coordinates for verices in UV_loops_stored with new values from uv_coords
		for loop_index in self.UV_loops_stored:
			index = me.loops[loop_index].vertex_index
			uvl[loop_index].uv[0] = uv_coords[match[index]][0]
			uvl[loop_index].uv[1] = uv_coords[match[index]][1]

		bpy.ops.object.mode_set(mode='EDIT')

# Initialize and register module 

def menu_draw(self, context):
	self.layout.operator("uv.island_copy", text="", icon="COPYDOWN")
	self.layout.operator("uv.island_paste", text="", icon="PASTEDOWN")
	
def register():
	bpy.utils.register_module(__name__)
	bpy.types.IMAGE_HT_header.append(menu_draw)
	
def unregister():
	bpy.types.IMAGE_HT_header.remove(menu_draw)
	bpy.utils.unregister_module(__name__)
	
if __name__ == "__main__":
	try:
		unregister()
	except:
		pass
	register()
